% This function implements restarted GMRES in matlab
%
function [x,res_norm,err_norm]=gmresrestart(l,u,a,f,reps,nrestart);
clear v hh cosgm singm rs x y xtrue 
n = size(a,1);
xtrue = a\f;
x = 0*f;
format long e

maxiter = 1000;
bnorm = norm(f);
itotal = 0;


r = f;
i = 0;
v(:,1) = r/norm(r);
rs(1) = norm(r);
res_norm(1) = rs(1);
rnorm = rs(1);
% disp(' ')
% disp(' ')
% disp('     iteration                 norm of the residual')
% disp('     ---------                 --------------------')
% disp(' ')
% disp(' ')

while (rnorm/bnorm) > reps & itotal < maxiter

r = f - a*x;
v(:,1) = r/norm(r);
rs(1) = norm(r);
i = 0;
rnorm = rs(1);

while (rnorm/bnorm) > reps & i < nrestart
   itotal = itotal+1;
   i = i+1;
   v(:,i+1) = a*(u\(l\v(:,i)));
   for j = 1:i
      hh(j,i) = v(:,j)'*v(:,i+1);
      v(:,i+1) = v(:,i+1) - hh(j,i)*v(:,j);
   end
   hh(i+1,i) = norm(v(:,i+1));
   v(:,i+1) = v(:,i+1)/hh(i+1,i);

%  --- Perform plane rotations on new column.

   if  i > 1 
      for k = 2: i
         thulp = hh(k-1,i);
         hh(k-1,i) =  cosgm(k-1)*thulp + singm(k-1)*hh(k,i);
         hh(k,i)   = -singm(k-1)*thulp + cosgm(k-1)*hh(k,i);
      end
   end

%  --- Calculate new plane rotation.

   thulp = sqrt(hh(i,i)^2 + hh(i+1,i)^2);
   cosgm(i)  = hh(i,i)/thulp;
   singm(i)  = hh(i+1,i)/thulp;
   rs(i+1)   = -singm(i)*rs(i);
   rs(i)     = cosgm(i)*rs(i);
   hh(i,i) = cosgm(i)*hh(i,i) + singm(i)*hh(i+1,i);
   res_norm(itotal+1) = abs(rs(i+1));
   rnorm = res_norm(itotal+1);
%    disp([itotal,res_norm(itotal+1)])
end

%  --- Compute new solution. First solve upper triangular system.

rs(i) = rs(i)/hh(i,i);
for k = i-1: -1: 1
   thulp = rs(k);
   for j = k+1: i
      thulp = thulp-hh(k,j)*rs(j);
   end
   rs(k) = thulp/hh(k,k);
end
y = 0*x;
for j = 1:i
   y = y + rs(j)*v(:,j);
end
x = x + u\(l\y);

end

err_norm = norm(x-xtrue);

% disp(' ')
% disp(' ')
% disp('  The exact norm of the residual')
% disp(norm(f-a*x))
% disp('  The norm of the error')
% disp(norm(xtrue-x))
